#BAH and TPH estimation for ANF
#John Hogland 9/10/2019

library(mgcv)
library(ggplot2)
library(rgdal)
library(sp)
library(spdep)
baseDir <- "C:\\Users\\jshogland\\Documents\\John\\projects\\UMGradSchool\\Project\\Papers\\Dissertation\\chapter3\\data"
setwd(baseDir)
source("C:\\Users\\jshogland\\Documents\\John\\projects\\UMGradSchool\\Project\\Papers\\Dissertation\\Rscripts\\Dissertation\\Dissertation\\jhLib.R")

#' Title createEnsembledGam
#'
#' @param frm = formula
#' @param df = data frame
#' @param fam = family default = gaussian()
#' @param nmdl = number of models default = 50
#' @param ptrain = percent of data used to train the model default = 0.75
#' @param kfact = keep factor for models: used to select models that have a similar RMSE training and testing datasets (default 1.25 * training RMSE) 
#'
#' @return vector of gam models and oob and training rmse
#' @export
#'
#' @examples
createEnsembleGam <- function(frm, df, fam = gaussian(), nmdl = 50, ptrain = 0.75, kfact = 20) {
    mdlV = list(length = nmdl)
    rmseV = vector(mode = "double", length = nmdl)
    rmseT = vector(mode = "double", length = nmdl)
    n = round(ptrain * nrow(df))
    mdlCnt = 0
    while (mdlCnt < nmdl) {
        sIndex = sample(nrow(df), n)
        tdf = df[sIndex,]
        vdf = df[-sIndex,]
        try({
            mdl = gam(frm, family = fam, data = tdf)

            pvlV = getPredictedValues(mdl, vdf) #predict(mdl, newdata = vdf)#add multinom estimates
            ovlV = getResponseValues(vdf, mdl) #vdf[all.vars(frm)[1]][[1]]#adjust for multinom estimates
            t_rmseV = getErrorEstimate(pvlV, ovlV, mdl) #sqrt(mean((pvlV - ovlV) ^ 2))

            pvlT = getPredictedValues(mdl, tdf) #predict(mdl, newdata = tdf)
            ovlT = getResponseValues(tdf, mdl) #tdf[all.vars(frm)[1]][[1]]
            t_rmseT = getErrorEstimate(pvlT, ovlT, mdl) #sqrt(mean((pvlT - ovlT) ^ 2))

                if (t_rmseV <= (t_rmseT * kfact)) {
                    mdlCnt = mdlCnt + 1
                    mdlV[[mdlCnt]] = mdl
                    rmseV[mdlCnt] = t_rmseV
                    rmseT[mdlCnt] = t_rmseT
                    print(paste("Found model ", mdlCnt, sep = ""))
            }
        }, silent = TRUE)
    }
    return(list(mdlV,rmseV,rmseT))
}
getTrainOBBAIC <- function(EGAM) {
    n = length(EGAM[[1]])
    svl = 0
    svl2 = 0
    for (m in 1:n) {
        mdl = EGAM[[1]][[m]]
        svl = svl + mdl$aic
        svl2 = svl2+ mdl$null.deviance
    }
    return(c(mean(EGAM[[3]]), mean(EGAM[[2]]), svl / n, svl2/n))
}
getPredictedValues <- function(md, df, t = "response") {
    
    pVls = predict(md, newdata = df, type = t)
    fm = md$family
    if (fm$family == "multinom") {
        pVls <- apply(pVls, 1, function(x) which(max(x) == x)[1]) - 1
    }
    if (fm$family == "binomial") {
        pVls <- as.integer(pVls>0.5)
    }
    return (pVls)
}
getResponseValues <- function(df, md) {
    fm = md$family
    f = md$formula
    if (fm$family == "multinom") {
        f = f[[1]]
    }
    return(df[all.vars(f)[1]][[1]])
}
getErrorEstimate <- function(p, o, md) {
    fm = md$family
    outVl = NULL
    if (fm$family == "multinom" | fm$family == "binomial") {
        outVl = 1-sum(p==o)/length(p)
    }
    else {
        outVl = sqrt(mean((p - o) ^ 2))
    }
    return(outVl)
}
#' Transform data
#'Transforms data using pca cor
#' @param df =  data frame
#' @param response = response variable
#' @param pred = vector of predictor variables
#'
#' @return list (data frame, pca
#' @export
#'
#' @examples
transformData <- function(df, response, pred) {
    frm = formula(paste("~", paste(pred, collapse = "+"), sep = ""))
    pca = princomp(frm,data=df,cor=TRUE)
    return(list(data.frame(resp=df[response][[1]],pca$scores),pca))
}
#' Predict Bagged Gam model values
#'
#' @param bGamMdl = list of models
#' @param df = new data data frame
#'
#' @return = data frame of mean predictions and standard errors
#' @export
#'
#' @examples
predictEnsembleGam <- function(bGamMdl, df, trunc = 0) {
    fm = bGamMdl[[1]]$family
    m = NULL
    s = NULL
    mdls = length(bGamMdl)
    n = nrow(df)
    if (fm$family == "multinom") {
        nlp = fm$nlp
        sm = matrix(rep(0,n*(nlp+1)),nrow = n, ncol = nlp +1)
        s2m = sm
        for (i in seq(mdls)) {
            mdl = bGamMdl[[i]]
            p = predict(mdl, df, type = "response")
            sm = sm + p
            s2m = s2m + p ^ 2
        }
        m = sm / mdls
        s = sqrt((s2m - ((sm ^ 2) / mdls)) / (mdls - 1))
        return(list(m, s))

    }
    else {
        sV = vector(mode = "double", length = n)
        s2V = vector(mode = "double", length = n)
        for (i in seq(mdls)) {
            mdl = bGamMdl[[i]]
            p = predict(mdl, df, type = "response")
            if (fm$family != "binomial") { 
                p = p ^ 2
            }
            sV = sV + p
            s2V = s2V + p ^ 2
        }
        m = sV / mdls
        s = sqrt((s2V - ((sV ^ 2) / mdls)) / (mdls - 1))
        return(list(m,s))
    }
    
}

runSim <- function(resp, pred, df, nmdl = 10, ptrain = 0.75, kfact = 10, fam = gaussian(), outMdl = "") {

    respVl = df[resp][[1]]
    nlp = 1
    if (fam$family == "multinom") {
        nlp = fam$nlp
    }
    fm <- getFormula(resp, pred, nlp) #as.formula(paste(resp, " ~ ", paste("s(", pred, ")", collapse = " + ")))
    mdls = createEnsembleGam(fm, df, nmdl = nmdl, ptrain = ptrain, kfact = kfact, fam = fam)
    if (outMdl != "") {
        saveRDS(mdls, outMdl)
    }
    p = predictEnsembleGam(mdls[[1]], df)
    est = p[[1]]
    lest = (est - p[[2]] * 1.96)
    uest = (est + p[[2]] * 1.96)

    if (fam$family == "multinom" | fam$family == "binomial") {
        pca = princomp(formula(paste("~", paste(pred, collapse = "+"), sep = "")), data = df, cor = TRUE)
        p1 = pca$scores[,1]
        if (fam$family == "multinom") {
            for (i in 1:ncol(est)) {
                fts = createGraph(p1, est[,i], resp, categorical = TRUE, pcol = respVl + 3)
            }
        }
        else {
            fts = createGraph(p1, est, resp, categorical = TRUE, pcol = respVl+3)
        }
    }
    else {
        fts = createGraph(respVl, est, resp)
    }
    

    #(rmse = sqrt(mean(res ^ 2)))
    #(r2 = 1 - ((mean(res ^ 2)) / var(respVl)))
    return(list(mdls[[1]], data.frame(TotRMSE = fts[2], TotR2 = fts[1] ^ 2, oobError = mean(mdls[[2]]), trainError = mean(mdls[[3]]))))
}

createGraph <- function(obs, est, graphName, axisTitleBlank = TRUE, categorical=FALSE,pcol="blue") {
    xTitle = "Observed"
    if (categorical==TRUE) {
        xTitle = "Comp_1"
    }
    yTitle = "Predicted"
    if (axisTitleBlank == FALSE) {
        xTitle = ""
        yTitle = ""
    }

    cr = cor(est, obs)
    rmse = sqrt(mean((obs - est) ^ 2))
    sbt = paste("Correlation = ",round(cr,digits=2),": RMSE = ", round(rmse,digits=2),sep = "")
    p4 = ggplot(data.frame(Observed = obs, Predicted = est), aes(x = Observed, y = Predicted)) +
        theme_bw() +
        geom_smooth(size = 1, linetype = "dotdash", col = "black") +
        geom_point(col = pcol, pch = 1) +
        geom_abline(slope = 1, intercept = 0, col = "red", size = 0.5) + 
        theme(legend.position = "none", plot.subtitle = element_text(size = 15), plot.title = element_text(size = 18), axis.text = element_text(size = 15), axis.title = element_text(size = 17)) +
        #coord_cartesian(ylim = c(0, 45), xlim = c(0, 45)) +
        
        xlab(xTitle) +
        ylab(yTitle) +
        
    labs(title = graphName,subtitle = sbt)
    print(p4)
    return(c(cr,rmse))
}

#get dataset and predictor names
plts = readOGR("PlotsShifted244.shp")
AnfDf = plts@data
knn = knearneigh(plts@coords, k = 1)
nb = knn2nb(knn)
AnfDf$DFT = 2
AnfDf[(AnfDf$Pine1BAH + AnfDf$Other1BAH) < 2, "DFT"] = 0
AnfDf[AnfDf$Other1BAH > AnfDf$Pine1BAH & AnfDf$DFT != 0, "DFT"] = 1
AnfDf$LPP = 0
AnfDf[AnfDf$LP1BAH >= 2, "LPP"] = 1
crd = plts@coords
AnfDf$X = crd[,1]
AnfDf$Y = crd[, 2]
AnfDf$XY = AnfDf$X * AnfDf$Y
(predLs <- names(AnfDf)[17:52])
(predSn <- names(AnfDf)[55:78])
(predNp <- names(AnfDf)[79:86])
(predXY <- names(AnfDf)[92:94])
(allPred <- c(predLs, predSn, predNp, predXY))
(allImagePred <- c(predLs, predSn, predNp))
tdf = transformData(AnfDf, "Pine1BAH", allImagePred)
pcaImage = tdf[[2]]
pcaDf = tdf[[1]]
(pcaPred = names(pcaDf)[2:length(names(pcaDf))])

#get dataset and predictor names for raw
AnfDfRaw = read.csv("ANFplotsRaw3.csv")
AnfDfRaw = subset(AnfDfRaw, !(OBJECTID == 189 | OBJECTID == 197 | OBJECTID == 204 | OBJECTID == 207))
#AnfVRaw = subset(AnfDfAllRaw, sample == 0)
AnfDfRaw$DFT = AnfDf$DFT
AnfDfRaw$LPP = AnfDf$LPP
AnfDfRaw$X = AnfDf$X
AnfDfRaw$Y = AnfDf$Y
AnfDfRaw$XY = AnfDf$XY
(predLsRaw <- names(AnfDfRaw)[22:57])
(predSnRaw <- names(AnfDfRaw)[58:81])
(predNpRaw <- names(AnfDfRaw)[82:89])
(predXYRaw <-  names(AnfDfRaw)[92:94])
(allPredRaw <- c(predLsRaw, predSnRaw, predNpRaw, predXYRaw))
(allImagePredRaw <- c(predLsRaw, predSnRaw, predNpRaw))
tdfRaw = transformData(AnfDfRaw, "Pine1BAH", allImagePredRaw)
pcaImageRaw = tdfRaw[[2]]
pcaDfRaw = tdfRaw[[1]]
(pcaPredRaw = names(pcaDfRaw)[2:length(names(pcaDfRaw))])

#getDFTModels
(resp = "DFT")
bmdl = getGamSigFldNames(AnfDf, resp, allImagePred, 0.05, fam = multinom(K = 2))
(pred = bmdl[[1]])
summary(bmdl[[2]])
DftStats = runSim(resp, pred, AnfDf, 50, ptrain = 0.75, kfact = 100,fam=multinom(K=2), outMdl = "Dft244.egm")
DftStats[[2]]
DftStats = readRDS("Dft244.egm")
getTrainOBBAIC(DftStats)
pDFT = predictEnsembleGam(DftStats[[1]], AnfDf)
AnfDf$r_pDFT0 = as.integer(AnfDf$DFT == 0) - pDFT[[1]][, 1]
AnfDf$r_pDFT1 = as.integer(AnfDf$DFT == 1) - pDFT[[1]][, 2]
AnfDf$r_pDFT2 = as.integer(AnfDf$DFT == 2) - pDFT[[1]][, 3]
(mt = moran.test(AnfDf$r_pDFT0, nb2listw(nb)))
(mt = moran.test(AnfDf$r_pDFT1, nb2listw(nb)))
(mt = moran.test(AnfDf$r_pDFT2, nb2listw(nb)))

(resp = "DFT")
bmdlRaw = getGamSigFldNames(AnfDfRaw, resp, allImagePredRaw, 0.05, fam = multinom(K = 2))
(predRaw = bmdlRaw[[1]])
summary(bmdlRaw[[2]])
DftStatsRaw = runSim(resp, predRaw, AnfDfRaw, 50, ptrain = 0.75, kfact = 100, fam=multinom(K=2),outMdl = "DftRaw244.egm")
DftStatsRaw[[2]]
DftStatsRaw = readRDS("DftRaw244.egm")
getTrainOBBAIC(DftStatsRaw)
pDFTRaw = predictEnsembleGam(DftStatsRaw[[1]], AnfDfRaw)
AnfDf$r_pDFT0 = as.integer(AnfDfRaw$DFT == 0) - pDFTRaw[[1]][, 1]
AnfDf$r_pDFT1 = as.integer(AnfDfRaw$DFT == 1) - pDFTRaw[[1]][, 2]
AnfDf$r_pDFT2 = as.integer(AnfDfRaw$DFT == 2) - pDFTRaw[[1]][, 3]
(mt = moran.test(AnfDfRaw$r_pDFT0, nb2listw(nb)))
(mt = moran.test(AnfDfRaw$r_pDFT1, nb2listw(nb)))
(mt = moran.test(AnfDfRaw$r_pDFT2, nb2listw(nb)))


#getLPPModels
(resp = "LPP")
bmdl = getGamSigFldNames(AnfDf, resp, allImagePred, 0.05, fam = binomial())
(pred = bmdl[[1]])
summary(bmdl[[2]])
LppStats = runSim(resp, pred, AnfDf, 50, ptrain = 0.75, kfact = 100, fam = binomial(), outMdl = "LPP244.egm")
LppStats = readRDS("LPP244.egm")
getTrainOBBAIC(LppStats)
pLPP = predictEnsembleGam(LppStats[[1]], AnfDf)[[1]]
AnfDf$r_pLPP0 = as.vector((1 - pLPP) - as.integer(AnfDf$LPP == 0))
AnfDf$r_pLPP1 = as.vector(pLPP - as.integer(AnfDf$LPP == 1))
(mt = moran.test(AnfDf$r_pLPP0, nb2listw(nb)))
(mt = moran.test(AnfDf$r_pLPP1, nb2listw(nb)))

(resp = "LPP")
bmdl = getGamSigFldNames(AnfDfRaw, resp, allImagePredRaw, 0.05, fam = binomial())
(predRaw = bmdl[[1]])
summary(bmdl[[2]])
LppStatsRaw = runSim(resp, predRaw, AnfDfRaw, 50, ptrain = 0.75, kfact = 10, fam=binomial(), outMdl = "LppRaw244.egm")
LppStatsRaw = readRDS("LppRaw244.egm")
getTrainOBBAIC(LppStatsRaw)
pLPPRaw = predictEnsembleGam(LppStatsRaw[[1]], AnfDfRaw)[[1]]
AnfDfRaw$r_pLPP0 = as.vector((1 - pLPPRaw) - as.integer(AnfDfRaw$LPP == 0))
AnfDfRaw$r_pLPP1 = as.vector(pLPPRaw - as.integer(AnfDfRaw$LPP == 1))
(mt = moran.test(AnfDfRaw$r_pLPP0, nb2listw(nb)))
(mt = moran.test(AnfDfRaw$r_pLPP1, nb2listw(nb)))


#getPineModels
#PineBAH
AnfDf$sqrt_Pine1BAH = sqrt(AnfDf$Pine1BAH)
(resp = "sqrt_Pine1BAH")
bmdl = getGamSigFldNames(AnfDf, resp, allImagePred, 0.05, fam = gaussian())
(pred = bmdl[[1]])
summary(bmdl[[2]])
PineBahStats = runSim(resp,pred,AnfDf,50,ptrain = 0.75,kfact = 10,outMdl = "PineBah244_n.egm")
PineBahStats = readRDS("PineBah244_n.egm")
getTrainOBBAIC(PineBahStats)
pBAH = predictEnsembleGam(PineBahStats[[1]],AnfDf)
AnfDf$r_pBAH = as.vector(pBAH[[1]]^2 - AnfDf$Pine1BAH)
(mt = moran.test(AnfDf$r_pBAH, nb2listw(nb)))

AnfDfRaw$sqrt_Pine1BAH = sqrt(AnfDfRaw$Pine1BAH)
(resp = "sqrt_Pine1BAH")
predRaw = getGamSigFldNames(AnfDfRaw, resp, allImagePredRaw, 0.05, fam = gaussian())
PineBahStatsRaw = runSim(resp, predRaw[[1]], AnfDfRaw, 50, ptrain = 0.75, kfact = 10, outMdl = "PineBah244Raw_n.egm")
PineBahStatsRaw = readRDS("PineBah244Raw_n.egm")
getTrainOBBAIC(PineBahStatsRaw)
pBAHRaw = predictEnsembleGam(PineBahStatsRaw[[1]], AnfDfRaw)
AnfDfRaw$r_pBAH = as.vector(pBAHRaw[[1]] ^ 2 - AnfDfRaw$Pine1BAH)
(mt = moran.test(AnfDfRaw$r_pBAH, nb2listw(nb)))

#PineTPH
AnfDf$sqrt_Pine1TPH = sqrt(AnfDf$Pine1TPH)
(resp = "sqrt_Pine1TPH")
predTph = getGamSigFldNames(AnfDf, resp, allImagePred, 0.05, fam = gaussian())
PineTphStats = runSim(resp, predTph[[1]], AnfDf, 50, ptrain = 0.75, kfact = 10, outMdl = "PineTph244_n.egm")
PineTphStats = readRDS("PineTPH244.egm")
getTrainOBBAIC(PineTphStats)
pTPH = predictEnsembleGam(PineTphStats[[1]], AnfDf)
AnfDf$r_pTPH = as.vector(pTPH[[1]] ^ 2 - AnfDf$Pine1TPH)
(mt = moran.test(AnfDf$r_pTPH, nb2listw(nb)))

AnfDfRaw$sqrt_Pine1TPH = sqrt(AnfDfRaw$Pine1TPH)
(resp = "sqrt_Pine1TPH")
predRaw = getGamSigFldNames(AnfDfRaw, resp, allImagePredRaw, 0.05, fam = gaussian())
PineTphStatsRaw = runSim(resp, predRaw[[1]], AnfDfRaw, 50, ptrain = 0.75, kfact = 10, outMdl = "PineTph244Raw_n.egm")
PineTphStatsRaw = readRDS("PineTPH244Raw_n.egm")
getTrainOBBAIC(PineTphStatsRaw)
pTPHRaw = predictEnsembleGam(PineTphStatsRaw[[1]], AnfDfRaw)
AnfDfRaw$r_pTPH = as.vector(pTPHRaw[[1]] ^ 2 - AnfDfRaw$Pine1TPH)
(mt = moran.test(AnfDfRaw$r_pTPH, nb2listw(nb)))


#getOtherModels
#OthereBAH
AnfDf$sqrt_Other1BAH = sqrt(AnfDf$Other1BAH)
(resp = "sqrt_Other1BAH")
pred = getGamSigFldNames(AnfDf, resp, allImagePred, 0.05, fam = gaussian())
OtherBahStats = runSim(resp, pred[[1]], AnfDf, 50, ptrain = 0.75, kfact = 10, outMdl = "OtherBah244_n.egm")
OtherBahStats = readRDS("OtherBah244_n.egm")
getTrainOBBAIC(OtherBahStats)
oBAH = predictEnsembleGam(OtherBahStats[[1]], AnfDf)
AnfDf$r_oBAH = as.vector(oBAH[[1]] ^ 2 - AnfDf$Other1BAH)
(mt = moran.test(AnfDf$r_oBAH, nb2listw(nb)))


AnfDfRaw$sqrt_Other1BAH = sqrt(AnfDfRaw$Other1BAH)
(resp = "sqrt_Other1BAH")
predRaw = getGamSigFldNames(AnfDfRaw, resp, allImagePredRaw, 0.05, fam = gaussian())[[1]]
OtherBahStatsRaw = runSim(resp, predRaw, AnfDfRaw, 50, ptrain = 0.75, kfact = 10, outMdl = "OtherBahRaw244_n.egm")
OtherBahStatsRaw = readRDS("OtherBahRaw244_n.egm")
getTrainOBBAIC(OtherBahStatsRaw)
oBAHRaw = predictEnsembleGam(OtherBahStatsRaw[[1]], AnfDfRaw)
AnfDfRaw$r_oBAH = as.vector(oBAHRaw[[1]] ^ 2 - AnfDfRaw$Other1BAH)
(mt = moran.test(AnfDfRaw$r_oBAH, nb2listw(nb)))

#OtherTPH
AnfDf$sqrt_Other1TPH = sqrt(AnfDf$Other1TPH)
(resp = "sqrt_Other1TPH")
pred = getGamSigFldNames(AnfDf, resp, allImagePred, 0.05, fam = gaussian())[[1]]
OtherTphStats = runSim(resp, pred, AnfDf, 50, ptrain = 0.75, kfact = 10, outMdl = "OtherTph244_n.egm")
OtherTphStats = readRDS("OtherTph244_n.egm")
getTrainOBBAIC(OtherTphStats)
oTPH = predictEnsembleGam(OtherTphStats[[1]], AnfDf)
AnfDf$r_oTPH = as.vector(oTPH[[1]] ^ 2 - AnfDf$Other1TPH)
(mt = moran.test(AnfDf$r_oTPH, nb2listw(nb)))

AnfDfRaw$sqrt_Other1TPH = sqrt(AnfDfRaw$Other1TPH)
(resp = "sqrt_Other1TPH")
predRaw = getGamSigFldNames(AnfDfRaw, resp, allImagePredRaw, 0.05, fam = gaussian())[[1]]
OtherTphStatsRaw = runSim(resp, predRaw, AnfDfRaw, 50, ptrain = 0.75, kfact = 10, outMdl = "OtherTphRaw244_n.egm")
OtherTphStatsRaw = readRDS("OtherTphRaw244_n.egm")
getTrainOBBAIC(OtherTphStatsRaw)
oTPHRaw = predictEnsembleGam(OtherTphStatsRaw[[1]], AnfDfRaw)
AnfDfRaw$r_oTPH = as.vector(oTPHRaw[[1]] ^ 2 - AnfDfRaw$Other1TPH)
(mt = moran.test(AnfDfRaw$r_oTPH, nb2listw(nb)))

#Make Rasters
library(raster)
library(rgdal)
