#Create raster surfaces from ensemble models
#John Hogland 9/10/2019

library(rgdal)
library(raster)
library(mgcv)

createDomain <- function(Pred, rndDf, plotDf, outName, k = 100) {
    rndLocSub = subset(rndDf, select = names(Pred))
    AnfDfSub = subset(plotDf, select = names(Pred))
    km = kmeans(rndLocSub, k)
    ucls = unique(km$cluster)
    vl = predict(km, AnfDfSub)
    uvl = unique(vl)
    msvl = ucls[!(ucls %in% uvl)]
    msdf = data.frame(id = msvl, v = rep(1, length(msvl)))
    beginCluster(8)
    krs = clusterR(Pred, predict, args = list(model = km, fun = predict.kmeans), progress = 'text')
    drs = clusterR(krs, subs, args = list(msdf), filename = outName, format = "GTiff", datatype = "INT1U", NAFlag = 255, progress = 'text', overwrite = TRUE)
    endCluster()
    return(drs)
}

predict.kmeans <- function(x, newdata) {
    return(apply(newdata, 1, function(r) which.min(colSums((t(x$centers) - r) ^ 2))))
}

predictEnsembleGam <- function(bGamMdl, data, trunc = 0) {
    fm = bGamMdl[[1]]$family
    m = NULL
    s = NULL
    mdls = length(bGamMdl)
    n = nrow(data)
    if (fm$family == "multinom") {
        nlp = fm$nlp
        sm = matrix(rep(0, n * (nlp + 1)), nrow = n, ncol = nlp + 1)
        s2m = sm
        for (i in seq(mdls)) {
            mdl = bGamMdl[[i]]
            p = predict(mdl, data, type = "response")
            sm = sm + p
            s2m = s2m + p ^ 2
        }
        m = sm / mdls
        s = sqrt((s2m - ((sm ^ 2) / mdls)) / (mdls - 1))
        return(cbind(m,s))

    }
    else {
        sV = vector(mode = "double", length = n)
        s2V = vector(mode = "double", length = n)
        for (i in seq(mdls)) {
            mdl = bGamMdl[[i]]
            p = (predict(mdl, data, type = "response"))
            if (fm$family != "binomial") {
                p = p ^ 2
            }
            p[p < trunc] = trunc
            sV = sV + p
            s2V = s2V + p ^ 2
        }
        m = sV / mdls
        s = sqrt((s2V - ((sV ^ 2) / mdls)) / (mdls - 1))
        return(cbind(m,s))
    }

}

baseDir <- "C:\\Users\\jshogland\\Documents\\John\\projects\\RESTORE\\Outputs3"
setwd(baseDir)
rasterOptions(tmpdir = paste(getwd(), "tmp", sep = "/"))
tmpDir(create = TRUE)

# read rasters
rndLoc = readOGR("RndLoc10000.shp", "RndLoc10000")
rndDf = rndLoc@data
plotLoc = readOGR("PlotsShifted244.shp", "PlotsShifted_244")
AnfDf = plotLoc@data
AOI = shapefile("AOI.shp")
ext = extent(AOI)
LSP = brick("C:\\Users\\jshogland\\Documents\\John\\projects\\RESTORE\\Predictors\\LS.tif")
xmin = 882300
xmax = xmin + 30 * ncol(LSP)
ymin = 731820
ymax = ymin + 30 * nrow(LSP)
newExt = extent(c(xmin, xmax, ymin, ymax))
LSP2 = crop(setExtent(LSP, newExt), ext)
SNP = crop(brick("C:\\Users\\jshogland\\Documents\\John\\projects\\RESTORE\\Predictors\\SN.tif"),ext)
ARP = crop(brick("C:\\Users\\jshogland\\Documents\\John\\projects\\RESTORE\\Predictors\\Aerial.tif"),ext)


#subset and combine layers

#DFT
#beginCluster(8)
#DftStats = readRDS("Dft244.egm")
P1 = LSP2[[c(9)]]
P2 = SNP[[c(1, 4, 6, 10)]]
P3 = LSP2[[c(2, 10)]]
P4 = ARP[[c(1)]]
Pred = stack(P1,P2,P3,P4)
names(Pred) <- c("LS_Band16", "SN_Band2", "SN_Band5", "SN_Band7", "SN_Band21", "LS_Band3", "LS_Band17", "NP_Band1")
#Dft = clusterR(Pred, predict, args = list(model = DftStats[[1]], fun = predictEnsembleGam, index = 1:6), filename = "DFT.tif", format = "GTiff", verbose = TRUE, datatype = "FLT4S",NAFlag=-9999, progress='text')
#endCluster()
DFT_D = createDomain(Pred,rndDf = rndDf,plotDf=AnfDf,"DFT_D.tif")

#LPP
#beginCluster(8)
#LppStats = readRDS("LPP244.egm")
P1 = LSP2[[c(1)]]
P2 = SNP[[c(2,3)]]
P3 = LSP2[[c(10)]]
P4 = ARP[[c(2,3)]]
Pred = stack(P1, P2, P3,P4)
names(Pred) <- c("LS_Band2", "SN_Band3", "SN_Band4", "LS_Band17", "NP_Band2", "NP_Band5")
#LPP = clusterR(Pred, predict, args = list(model = LppStats[[1]], fun = predictEnsembleGam, index = 1:2), filename = "LPP.tif", format = "GTiff", verbose = TRUE, datatype = "FLT4S", NAFlag = -9999, progress = 'text')
#endCluster()
LPP_D = createDomain(Pred, rndDf = rndDf, plotDf = AnfDf, "LPP_D.tif")

#baseDir <- "E:\\Projects\\RESTORE\\Outputs2"
#setwd(baseDir)
#rasterOptions(tmpdir = paste(getwd(), "tmp", sep = "/"))
#tmpDir(create = TRUE)

#PBAH
#beginCluster(8)
(P1 = LSP2[[c(1, 4, 6, 7, 9, 10, 11, 12)]])
(P2 = ARP[[c(2, 4)]])
(Pred = stack(P1, P2))
names(Pred)=c("LS_Band2","LS_Band8","LS_Band10","LS_Band11","LS_Band16","LS_Band17","LS_Band18","LS_Band22","NP_Band2","NP_Band8")
#PineBahStats = readRDS("PineBah244_n.egm")
#PBAH = clusterR(Pred, predict, args = list(model = PineBahStats[[1]], fun = predictEnsembleGam, index = 1:2),filename = "pBAH.tif",format="GTiff",verbose=TRUE,datatype="FLT4S", NAFlag = -9999, progress = 'text')
#endCluster()
PBAH_D = createDomain(Pred, rndDf = rndDf, plotDf = AnfDf, "PBAH_D.tif")

#PTPH
#beginCluster(8)
#PineTphStats = readRDS("PineTph244.egm")
(P1 = LSP2[[c(2,3,4,6,7,10,13)]])
(P2 = SNP[[c(3, 8)]])
(P3 = ARP[[c(2,3)]])
(Pred = stack(P1, P2, P3))
names(Pred) = c("LS_Band3", "LS_Band5", "LS_Band8", "LS_Band10", "LS_Band11", "LS_Band17", "LS_Band28", "SN_Band4", "SN_Band11", "NP_Band2", "NP_Band5")
#PTPH = clusterR(Pred, predict, args = list(model = PineTphStats[[1]], fun = predictEnsembleGam, index = 1:2), filename = "pTPH.tif", format = "GTiff", verbose = TRUE, datatype = "FLT4S", NAFlag = -9999, progress = 'text')
#endCluster()
PTPH_D = createDomain(Pred, rndDf = rndDf, plotDf = AnfDf, "PTPH_D.tif")


#OBAH
#beginCluster(8)
#OtherBahStats = readRDS("OtherBah244_n.egm")
(P1 = LSP2[[c(2, 3, 7)]])
(P2 = SNP[[c(6,7,10,11)]])
(P3 = ARP[[c(2, 3)]])
(Pred = stack(P1, P2, P3))
names(Pred) = c("LS_Band3", "LS_Band5", "LS_Band11", "SN_Band7", "SN_Band9", "SN_Band21", "SN_Band23", "NP_Band2", "NP_Band5")
#OBAH = clusterR(Pred, predict, args = list(model = OtherBahStats[[1]], fun = predictEnsembleGam, index = 1:2), filename = "oBAH.tif", format = "GTiff", verbose = TRUE, datatype = "FLT4S", NAFlag = -9999, progress = 'text')
#endCluster()
OBAH_D=createDomain(Pred, rndDf = rndDf, plotDf = AnfDf, "OBAH_D.tif")


#OTPH
#beginCluster(8)
#OtherTphStats = readRDS("OtherTph244_n.egm")
(P1 = LSP2[[c(5,8)]])
(P2 = SNP[[c(1,5,6,9)]])
(P3 = ARP[[c(1)]])
(Pred = stack(P1, P2, P3))
names(Pred) = c("LS_Band9", "LS_Band13", "SN_Band2", "SN_Band6", "SN_Band7", "SN_Band20", "NP_Band1")
#OTPH = clusterR(Pred, predict, args = list(model = OtherTphStats[[1]], fun = predictEnsembleGam, index = 1:2), filename = "oTPH.tif", format = "GTiff", verbose = TRUE, datatype = "FLT4S", NAFlag = -9999, progress = 'text')
#endCluster()
OTPH_D=createDomain(Pred, rndDf = rndDf, plotDf = AnfDf, "OTPH_D.tif")

#Create feature domain mask




removeTmpFiles(h = 0)